# Copyright (c) OpenMMLab. All rights reserved.

import torch
import torch.nn.functional as F
from mmpose.models.utils import rope

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.utils.rtmcc_block.ScaleNorm.forward", backend="ncnn"
)
def scalenorm__forward__ncnn(self, x):
    """Rewrite `scalenorm` for ncnn backend.

    Rewrite scalenorm to avoid FP16 exceed in ncnn Android platform.
    """
    # The one-dim of Fubinious norm is equal to L2Norm.
    # Set p=2 explicitly to map torch.norm to ReduceL2 onnx op,
    # which will avoid FP16 exceed.
    norm = torch.norm(x, dim=2, keepdim=True)
    norm = norm * self.scale
    # Rewrite for ncnn binaryop broadcast.
    norm = norm.clamp(min=self.eps)
    return (x.unsqueeze(2) / norm.unsqueeze(2)).squeeze(2) * self.g


@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.utils.rtmcc_block.RTMCCBlock._forward", backend="ncnn"
)
def rtmccblock___forward_ncnn(self, inputs):
    """Rewrite `_forward` of RTMBlock for ncnn backend.

    Rewrite the matmul and avoid unbind for ncnn backend.
    """
    if self.attn_type == "self-attn":
        x = inputs
    else:
        x, k, v = inputs

    x = self.ln(x)
    uv = self.uv(x)
    if self.attn_type == "self-attn":
        uv = self.act_fn(uv)
        u = uv[..., : self.e]
        v = uv[..., self.e : 2 * self.e]
        base = uv[..., 2 * self.e : 2 * self.e + self.s]

        q = (
            base.unsqueeze(1) * self.gamma[None, None, 0:1, :]
            + self.beta[None, None, 0:1, :]
        ).squeeze(1)
        k = (
            base.unsqueeze(1) * self.gamma[None, None, 1:2, :]
            + self.beta[None, None, 1:2, :]
        ).squeeze(1)

        if self.pos_enc:
            q = rope(q, dim=1)
            k = rope(k, dim=1)
    else:
        u, q = torch.split(self.act_fn(uv), [self.e, self.s], dim=-1)

        k = self.k_fc(k)
        v = self.v_fc(v)

        if self.pos_enc:
            q = rope(q, 1)
            k = rope(k, 1)
    qk = torch.bmm(q, k.permute(0, 2, 1))
    if self.use_rel_bias:
        if self.attn_type == "self-attn":
            bias = self.rel_pos_bias(q.size(1))
        else:
            bias = self.rel_pos_bias(q.size(1), k.size(1))
        qk += bias[:, : q.size(1), : k.size(1)]

    kernel = torch.square(F.relu(qk / self.sqrt_s))
    if self.dropout_rate > 0.0:
        kernel = self.dropout(kernel)

    x = u * torch.bmm(kernel, v)
    x = self.o(x)

    return x


@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.utils.rtmcc_block.Scale.forward", backend="ncnn"
)
def scale__forward_ncnn(self, x):
    """Rewrite `forward` of Scale for ncnn backend.

    Adapt the shape to avoid ncnn BinaryOp seg fault.
    """
    x = x.unsqueeze(1)
    scale = self.scale[None, None, None, :]
    return (x * scale).squeeze(1)



# Copyright (c) OpenMMLab. All rights reserved.

import torch.nn.functional as F
from torch.nn.modules.utils import _pair

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_root_logger, is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.adaptive_avg_pool2d"
)
def adaptive_avg_pool2d__default(input, output_size):
    """Rewrite `adaptive_avg_pool2d` for default backend."""
    ctx = FUNCTION_REWRITER.get_context()
    output_size = _pair(output_size)
    if int(output_size[0]) == int(output_size[1]) == 1:
        out = ctx.origin_func(input, output_size)
    else:
        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        if is_dynamic_flag:
            logger = get_root_logger()
            logger.warning(
                "`adaptive_avg_pool2d` would be " "replaced to `avg_pool2d` explicitly"
            )
        size = input.shape[2:]
        k = [int(size[i] / output_size[i]) for i in range(0, len(size))]
        out = F.avg_pool2d(
            input,
            kernel_size=k,
            stride=k,
            padding=0,
            ceil_mode=False,
            count_include_pad=False,
        )
    return out


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.adaptive_avg_pool2d", backend=Backend.NCNN.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.adaptive_avg_pool2d",
    backend=Backend.TORCHSCRIPT.value,
)
def adaptive_avg_pool2d__ncnn(input, output_size):
    ctx = FUNCTION_REWRITER.get_context()
    """Rewrite `adaptive_avg_pool2d` for ncnn and torchscript backend."""
    return ctx.origin_func(input, output_size)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.flatten", backend=Backend.NCNN.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.flatten", backend=Backend.NCNN.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.flatten", backend=Backend.COREML.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.flatten", backend=Backend.COREML.value
)
def flatten__coreml(input, start_dim=0, end_dim=-1) -> torch.Tensor:
    """Rewrite `flatten` for coreml backend.

    Use reshape instead of flatten
    """
    shape = input.shape
    end_dim = end_dim if end_dim > 0 else len(shape) + end_dim
    shape1 = list(shape[:start_dim])
    shape3 = list(shape[end_dim + 1 :])
    return input.reshape(shape1 + [-1] + shape3)



# Copyright (c) OpenMMLab. All rights reserved.
from . import base_detr, single_stage, single_stage_instance_seg, two_stage

__all__ = [
    'base_detr', 'single_stage', 'single_stage_instance_seg', 'two_stage'
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmdet.models.detectors.base import ForwardResults
from mmdet.structures.det_data_sample import OptSampleList

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.utils import is_dynamic_shape
from .single_stage import _set_metainfo


@mark(
    "instance_segmentor_forward", inputs=["input"], outputs=["dets", "labels", "masks"]
)
def __forward_impl_instance_seg(self, batch_inputs, data_samples, **kwargs):
    """Rewrite and adding mark for `forward`.

    Encapsulate this function for rewriting `forward` of BaseDetector.
    1. Add mark for BaseDetector.
    2. Support both dynamic and static export to onnx.
    """
    x = self.extract_feat(batch_inputs)
    mask_outs = self.mask_head.predict(x, data_samples, rescale=False)
    return mask_outs


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.detectors.single_stage_instance_seg."
    "SingleStageInstanceSegmentor.forward"
)
def single_stage_instance_segmentor__forward(
    self,
    batch_inputs: torch.Tensor,
    data_samples: OptSampleList = None,
    mode: str = "tensor",
    **kwargs,
) -> ForwardResults:
    """Rewrite `forward` for default backend.
    Support configured dynamic/static shape for model input and return
    detection result as Tensor instead of numpy array.
    Args:
        batch_inputs (Tensor): Inputs with shape (N, C, H, W).
        data_samples (List[:obj:`DetDataSample`]): The Data
            Samples. It usually includes information such as
            `gt_instance`, `gt_panoptic_seg` and `gt_sem_seg`.
        rescale (bool): Whether to rescale the results.
            Defaults to True.
    Returns:
        tuple[Tensor]: Detection results of the
        input images.
            - dets (Tensor): Classification bboxes and scores.
                Has a shape (num_instances, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg

    # get origin input shape as tensor to support onnx dynamic shape
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    img_shape = torch._shape_as_tensor(batch_inputs)[2:]
    if not is_dynamic_flag:
        img_shape = [int(val) for val in img_shape]

    # set the metainfo
    data_samples = _set_metainfo(data_samples, img_shape)

    return __forward_impl_instance_seg(
        self, batch_inputs, data_samples=data_samples, **kwargs
    )



_base_ = ["./pose-detection_static.py", "../_base_/backends/onnxruntime.py"]

onnx_config = dict(
    input_shape=[192, 256],
    output_names=["simcc_x", "simcc_y"],
    dynamic_axes={
        "input": {
            0: "batch",
        },
        "simcc_x": {0: "batch"},
        "simcc_y": {0: "batch"},
    },
)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.utils.transformer.PatchMerging.forward", backend="tensorrt"
)
def patch_merging__forward__tensorrt(self, x, input_size):
    """Rewrite forward function of PatchMerging class for TensorRT. In original
    implementation, mmdet applies nn.unfold to accelerate the inference.
    However, the onnx graph of it can not be parsed correctly by TensorRT. In
    mmdeploy, it is replaced.

    Args:
        x (Tensor): Has shape (B, H*W, C_in).
        input_size (tuple[int]): The spatial shape of x, arrange as (H, W).
            Default: None.

    Returns:
        tuple: Contains merged results and its spatial shape.
            - x (Tensor): Has shape (B, Merged_H * Merged_W, C_out)
            - out_size (tuple[int]): Spatial shape of x, arrange as
                (Merged_H, Merged_W).
    """
    H, W = input_size
    B, L, C = x.shape
    assert L == H * W, "input feature has wrong size"
    assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

    x = x.view(B, H, W, C)

    x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
    x1 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
    x2 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
    x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
    x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
    x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C
    x = (
        x.view(x.shape[0], x.shape[1], 4, -1)
        .permute(0, 1, 3, 2)
        .reshape(x.shape[0], x.shape[1], -1)
    )
    x = self.norm(x) if self.norm else x
    x = self.reduction(x)
    out_h = (
        H
        + 2 * self.sampler.padding[0]
        - self.sampler.dilation[0] * (self.sampler.kernel_size[0] - 1)
        - 1
    ) // self.sampler.stride[0] + 1
    out_w = (
        W
        + 2 * self.sampler.padding[1]
        - self.sampler.dilation[1] * (self.sampler.kernel_size[1] - 1)
        - 1
    ) // self.sampler.stride[1] + 1

    output_size = (out_h, out_w)
    return x, output_size



# Copyright (c) OpenMMLab. All rights reserved.

from juxtapose.mmdeploy.backend.ascend import is_available

__all__ = ["is_available"]

if is_available():
    from juxtapose.mmdeploy.backend.ascend.onnx2ascend import from_onnx as _from_onnx
    from ..core import PIPELINE_MANAGER

    from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)
    __all__ += ["from_onnx"]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from torch import Tensor

import juxtapose.mmdeploy
from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark


class ONNXNMSRotatedOp(torch.autograd.Function):
    """Create onnx::NMSRotated op."""

    @staticmethod
    def forward(
        ctx, boxes: Tensor, scores: Tensor, iou_threshold: float, score_threshold: float
    ) -> Tensor:
        """Get NMS rotated output indices.

        Args:
            ctx (Context): The context with meta information.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 5].
            scores (Tensor): The detection scores of shape
                [N, num_classes, num_boxes].
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): bbox threshold, bboxes with scores
            lower than it will not be considered.

        Returns:
            Tensor: Selected indices of boxes.
        """
        from mmcv.ops import nms_rotated

        batch_size, num_class, _ = scores.shape

        indices = []
        for batch_id in range(batch_size):
            for cls_id in range(num_class):
                _boxes = boxes[batch_id, ...]
                # score_threshold=0 requires scores to be contiguous
                _scores = scores[batch_id, cls_id, ...].contiguous()
                valid_mask = _scores > score_threshold
                _boxes, _scores = _boxes[valid_mask], _scores[valid_mask]
                if _boxes.shape[0] == 0:
                    continue
                valid_inds = torch.nonzero(valid_mask, as_tuple=False).squeeze(dim=1)
                _, box_inds = nms_rotated(_boxes, _scores, iou_threshold=iou_threshold)
                box_inds = valid_inds[box_inds]
                batch_inds = torch.zeros_like(box_inds) + batch_id
                cls_inds = torch.zeros_like(box_inds) + cls_id
                indices.append(torch.stack([batch_inds, cls_inds, box_inds], dim=-1))

        indices = (
            torch.cat(indices)
            if len(indices) > 0
            else torch.zeros((0, 3), dtype=torch.long, device=boxes.device)
        )
        return indices

    @staticmethod
    def symbolic(
        g, boxes: Tensor, scores: Tensor, iou_threshold: float, score_threshold: float
    ):
        """Symbolic function for onnx::NMSRotated.

        Args:
            g (Graph): The traced onnx graph.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): bbox threshold, bboxes with scores
            lower than it will not be considered.

        Returns:
            NMSRotated op for onnx.
        """
        return g.op(
            "mmdeploy::NMSRotated",
            boxes,
            scores,
            iou_threshold_f=float(iou_threshold),
            score_threshold_f=float(score_threshold),
        )


class TRTBatchedRotatedNMSop(torch.autograd.Function):
    """Create mmdeploy::TRTBatchedRotatedNMSop op for TensorRT backend.

    NMS in ONNX supports dynamic outputs. This class helps replace
    onnx::NonMaxSuppression with mmdeploy::TRTBatchedRotatedNMSop.
    """

    @staticmethod
    def forward(
        ctx,
        boxes: Tensor,
        scores: Tensor,
        num_classes: int,
        pre_topk: int,
        after_topk: int,
        iou_threshold: float,
        score_threshold: float,
        background_label_id: int = -1,
    ):
        """Forward of batched rotated nms.

        Args:
            ctx (Context): The context with meta information.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 5].
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            num_classes (int): MThe number of classes in the network.
            pre_topk (int): The number of bounding boxes to be fed into
                the NMS step.
            after_topk (int): The number of total bounding boxes to be kept
                per-image after the NMS step. Should be less than or equal
                to the pre_topk value.
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): score threshold of nms.
            background_label_id (int): The label ID for the background class.
                If there is no background class, set it to -1.

        Returns:
            dets (Tensor): Bboxes and scores of the rotated nms results.
            labels (Tensor): Class id of the rotated nms results.
        """
        batch_size, num_boxes, num_classes = scores.shape

        out_boxes = min(num_boxes, after_topk)
        return torch.rand(batch_size, out_boxes, 6).to(scores.device), torch.randint(
            0, num_classes, (batch_size, out_boxes)
        ).to(scores.device)

    @staticmethod
    def symbolic(
        g,
        boxes: Tensor,
        scores: Tensor,
        num_classes: int,
        pre_topk: int,
        after_topk: int,
        iou_threshold: float,
        score_threshold: float,
        background_label_id: int = -1,
    ):
        """Symbolic function for mmdeploy::TRTBatchedNMS."""
        return g.op(
            "mmdeploy::TRTBatchedRotatedNMS",
            boxes,
            scores,
            num_classes_i=num_classes,
            background_label_id_i=background_label_id,
            iou_threshold_f=iou_threshold,
            score_threshold_f=score_threshold,
            topk_i=pre_topk,
            keep_topk_i=after_topk,
            is_normalized_i=False,
            clip_boxes_i=False,
            outputs=2,
        )


def select_rnms_index(
    scores: torch.Tensor,
    boxes: torch.Tensor,
    nms_index: torch.Tensor,
    batch_size: int,
    keep_top_k: int = -1,
):
    """Transform NMSRotated output.

    Args:
        scores (Tensor): The detection scores of shape
            [N, num_classes, num_boxes].
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 6].
        nms_index (Tensor): NMS output of bounding boxes indexing.
        batch_size (int): Batch size of the input image.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 6]
            and `labels` of shape [N, num_det].
    """
    batch_inds, cls_inds = nms_index[:, 0], nms_index[:, 1]
    box_inds = nms_index[:, 2]

    # index by nms output
    scores = scores[batch_inds, cls_inds, box_inds].unsqueeze(1)
    boxes = boxes[batch_inds, box_inds, ...]
    dets = torch.cat([boxes, scores], dim=1)

    # batch all
    batched_dets = dets.unsqueeze(0).repeat(batch_size, 1, 1)
    batch_template = torch.arange(
        0, batch_size, dtype=batch_inds.dtype, device=batch_inds.device
    )
    batched_dets = batched_dets.where(
        (batch_inds == batch_template.unsqueeze(1)).unsqueeze(-1),
        batched_dets.new_zeros(1),
    )

    batched_labels = cls_inds.unsqueeze(0).repeat(batch_size, 1)
    batched_labels = batched_labels.where(
        (batch_inds == batch_template.unsqueeze(1)), batched_labels.new_ones(1) * -1
    )

    N = batched_dets.shape[0]

    # expand tensor to eliminate [0, ...] tensor
    batched_dets = torch.cat((batched_dets, batched_dets.new_zeros((N, 1, 6))), 1)
    batched_labels = torch.cat((batched_labels, batched_labels.new_zeros((N, 1))), 1)

    # sort
    is_use_topk = keep_top_k > 0 and (
        torch.onnx.is_in_onnx_export() or keep_top_k < batched_dets.shape[1]
    )
    if is_use_topk:
        _, topk_inds = batched_dets[:, :, -1].topk(keep_top_k, dim=1)
    else:
        _, topk_inds = batched_dets[:, :, -1].sort(dim=1, descending=True)
    topk_batch_inds = torch.arange(
        batch_size, dtype=topk_inds.dtype, device=topk_inds.device
    ).unsqueeze(1)
    batched_dets = batched_dets[topk_batch_inds, topk_inds, ...]
    batched_labels = batched_labels[topk_batch_inds, topk_inds, ...]

    # slice and recover the tensor
    return batched_dets, batched_labels


def _multiclass_nms_rotated(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.1,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
):
    """NMSRotated for multi-class bboxes.

    This function helps exporting to onnx with batch and multiclass NMSRotated
    op. It only supports class-agnostic detection results. That is, the scores
    is of shape (N, num_bboxes, num_classes) and the boxes is of shape
    (N, num_boxes, 5).

    Args:
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 5].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): bbox threshold, bboxes with scores lower than
            it will not be considered.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 6]
            and `labels` of shape [N, num_det].
    """
    batch_size = scores.shape[0]

    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.topk(pre_top_k)
        batch_inds = torch.arange(batch_size).unsqueeze(1).long()
        boxes = boxes[batch_inds, topk_inds, :]
        scores = scores[batch_inds, topk_inds, :]

    scores = scores.permute(0, 2, 1)
    selected_indices = ONNXNMSRotatedOp.apply(
        boxes, scores, iou_threshold, score_threshold
    )

    dets, labels = select_rnms_index(
        scores, boxes, selected_indices, batch_size, keep_top_k=keep_top_k
    )

    return dets, labels


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.mmcv.ops.nms_rotated._multiclass_nms_rotated",
    backend="tensorrt",
)
def multiclass_nms_rotated__tensorrt(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
):
    """Wrapper for `multiclass_nms` with TensorRT.

    Args:
        ctx (ContextCaller): The context with additional information.
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 5].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 6]
            and `labels` of shape [N, num_det].
    """
    boxes = boxes if boxes.dim() == 4 else boxes.unsqueeze(2)
    keep_top_k = (
        max_output_boxes_per_class
        if keep_top_k < 0
        else min(max_output_boxes_per_class, keep_top_k)
    )
    dets, labels = TRTBatchedRotatedNMSop.apply(
        boxes,
        scores,
        int(scores.shape[-1]),
        pre_top_k,
        keep_top_k,
        iou_threshold,
        score_threshold,
        -1,
    )

    return dets, labels


@mark("multiclass_nms_rotated", inputs=["boxes", "scores"], outputs=["dets", "labels"])
def multiclass_nms_rotated(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.1,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
):
    """Wrapper function for `_multiclass_nms`."""
    return juxtapose.mmdeploy.mmcv.ops.nms_rotated._multiclass_nms_rotated(
        boxes,
        scores,
        max_output_boxes_per_class=max_output_boxes_per_class,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict

import mmengine
import torch.nn as nn

from juxtapose.mmdeploy.utils.constants import IR, Backend
from .function_rewriter import FunctionRewriter
from .module_rewriter import ModuleRewriter
from .rewriter_utils import collect_env
from .symbolic_rewriter import SymbolicRewriter


class RewriterManager:
    """The rewrite manager that manages some rewriters."""

    def __init__(self):
        self.module_rewriter = ModuleRewriter()
        self.function_rewriter = FunctionRewriter()
        self.symbolic_rewriter = SymbolicRewriter()


REWRITER_MANAGER = RewriterManager()

MODULE_REWRITER = REWRITER_MANAGER.module_rewriter
FUNCTION_REWRITER = REWRITER_MANAGER.function_rewriter
SYMBOLIC_REWRITER = REWRITER_MANAGER.symbolic_rewriter


def patch_model(
    model: nn.Module,
    cfg: mmengine.Config,
    backend: str = Backend.DEFAULT.value,
    ir: IR = IR.DEFAULT,
    recursive: bool = True,
    **kwargs,
) -> nn.Module:
    """Patch the model, replace the modules that can be rewritten. Note that
    the original model will be modified permanently.

    Args:
        model (torch.nn.Module): The model to patch.
        cfg (Dict): Config dictionary of deployment.
        backend (str): The inference engine name.
        ir (IR): The intermeditate representation name.
        recursive (bool): The flag to enable recursive patching.

    Returns:
        nn.Module: THe patched model.

    Examples:
        >>> from juxtapose.mmdeploy.core import patch_model
        >>> from juxtapose.mmdeploy.utils import Backend, IR
        >>> deploy_cfg = {}
        >>> backend = Backend.DEFAULT.value
        >>> ir = IR.ONNX
        >>> patched_model = patch_model(model, deploy_cfg, backend, ir)
    """
    return MODULE_REWRITER.patch_model(model, cfg, backend, ir, recursive, **kwargs)


class RewriterContext:
    """The rewrite context.

    The context is used to manage the rewrite functions and the backend.

    Args:
        cfg (Dict): Config dictionary of deployment.
        backend (str): The inference engine name.
        ir (IR): The intermeditate representation name.
        rewrite_manager (RewriterManager): An RewriteManager that consists of
            several rewriters

    Examples:
        >>> from juxtapose.mmdeploy.core import RewriterContext
        >>> with RewriterContext(cfg, backend='onnxruntime'):
        >>>     # the rewrite has been activated inside the context
        >>>     torch.onnx.export(model, inputs, onnx_file)
    """

    def __init__(
        self,
        cfg: Dict = dict(),
        backend: str = Backend.DEFAULT.value,
        ir: IR = IR.DEFAULT,
        rewriter_manager: RewriterManager = REWRITER_MANAGER,
        **kwargs,
    ):
        self._cfg = cfg
        self._kwargs = kwargs
        self._rewriter_manager = rewriter_manager
        self._env = collect_env(Backend.get(backend), ir)

    def enter(self):
        """Call the enter() of rewriters."""
        self._rewriter_manager.function_rewriter.enter(
            self._cfg, self._env, **self._kwargs
        )
        self._rewriter_manager.symbolic_rewriter.enter(
            self._cfg, self._env, **self._kwargs
        )

    def exit(self):
        """Call the exit() of rewriters."""
        self._rewriter_manager.function_rewriter.exit()
        self._rewriter_manager.symbolic_rewriter.exit()

    def __enter__(self):
        """Call enter()"""
        self.enter()

    def __exit__(self, type, val, tb):
        """Call exit()"""
        self.exit()



# Copyright (c) OpenMMLab. All rights reserved.
from . import adaptive_pool  # noqa: F401,F403
from . import gelu  # noqa: F401,F403
from . import grid_sampler  # noqa: F401,F403
from . import hardsigmoid  # noqa: F401,F403
from . import instance_norm  # noqa: F401,F403
from . import layer_norm  # noqa: F401,F403
from . import linear  # noqa: F401,F403
from . import lstm  # noqa: F401,F403
from . import roll  # noqa: F401,F403
from . import squeeze  # noqa: F401,F403



_base_ = ['mmpose::_base_/default_runtime.py']

# runtime
max_epochs = 420
stage2_num_epochs = 30
base_lr = 4e-3

train_cfg = dict(max_epochs=max_epochs, val_interval=10)
randomness = dict(seed=21)

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='AdamW', lr=base_lr, weight_decay=0.05),
    paramwise_cfg=dict(
        norm_decay_mult=0, bias_decay_mult=0, bypass_duplicate=True))

# learning rate
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0e-5,
        by_epoch=False,
        begin=0,
        end=1000),
    dict(
        # use cosine lr from 210 to 420 epoch
        type='CosineAnnealingLR',
        eta_min=base_lr * 0.05,
        begin=max_epochs // 2,
        end=max_epochs,
        T_max=max_epochs // 2,
        by_epoch=True,
        convert_to_iter_based=True),
]

# automatically scaling LR based on the actual training batch size
auto_scale_lr = dict(base_batch_size=1024)

# codec settings
codec = dict(
    type='SimCCLabel',
    input_size=(192, 256),
    sigma=(4.9, 5.66),
    simcc_split_ratio=2.0,
    normalize=False,
    use_dark=False)

# model settings
model = dict(
    type='TopdownPoseEstimator',
    data_preprocessor=dict(
        type='PoseDataPreprocessor',
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        bgr_to_rgb=True),
    backbone=dict(
        _scope_='mmdet',
        type='CSPNeXt',
        arch='P5',
        expand_ratio=0.5,
        deepen_factor=1.,
        widen_factor=1.,
        out_indices=(4, ),
        channel_attention=True,
        norm_cfg=dict(type='SyncBN'),
        act_cfg=dict(type='SiLU'),
        init_cfg=dict(
            type='Pretrained',
            prefix='backbone.',
            checkpoint='https://download.openmmlab.com/mmpose/v1/projects/'
            'rtmposev1/cspnext-l_udp-aic-coco_210e-256x192-273b7631_20230130.pth'  # noqa
        )),
    head=dict(
        type='RTMCCHead',
        in_channels=1024,
        out_channels=17,
        input_size=codec['input_size'],
        in_featuremap_size=(6, 8),
        simcc_split_ratio=codec['simcc_split_ratio'],
        final_layer_kernel_size=7,
        gau_cfg=dict(
            hidden_dims=256,
            s=128,
            expansion_factor=2,
            dropout_rate=0.,
            drop_path=0.,
            act_fn='SiLU',
            use_rel_bias=False,
            pos_enc=False),
        loss=dict(
            type='KLDiscretLoss',
            use_target_weight=True,
            beta=10.,
            label_softmax=True),
        decoder=codec),
    test_cfg=dict(flip_test=True))

# base dataset settings
dataset_type = 'CocoDataset'
data_mode = 'topdown'
data_root = 'data/coco/'

backend_args = dict(backend='local')
# backend_args = dict(
#     backend='petrel',
#     path_mapping=dict({
#         f'{data_root}': 's3://openmmlab/datasets/detection/coco/',
#         f'{data_root}': 's3://openmmlab/datasets/detection/coco/'
#     }))

# pipelines
train_pipeline = [
    dict(type='LoadImage', backend_args=backend_args),
    dict(type='GetBBoxCenterScale'),
    dict(type='RandomFlip', direction='horizontal'),
    dict(type='RandomHalfBody'),
    dict(
        type='RandomBBoxTransform', scale_factor=[0.6, 1.4], rotate_factor=80),
    dict(type='TopdownAffine', input_size=codec['input_size']),
    dict(type='mmdet.YOLOXHSVRandomAug'),
    dict(
        type='Albumentation',
        transforms=[
            dict(type='Blur', p=0.1),
            dict(type='MedianBlur', p=0.1),
            dict(
                type='CoarseDropout',
                max_holes=1,
                max_height=0.4,
                max_width=0.4,
                min_holes=1,
                min_height=0.2,
                min_width=0.2,
                p=1.),
        ]),
    dict(type='GenerateTarget', encoder=codec),
    dict(type='PackPoseInputs')
]
val_pipeline = [
    dict(type='LoadImage', backend_args=backend_args),
    dict(type='GetBBoxCenterScale'),
    dict(type='TopdownAffine', input_size=codec['input_size']),
    dict(type='PackPoseInputs')
]

train_pipeline_stage2 = [
    dict(type='LoadImage', backend_args=backend_args),
    dict(type='GetBBoxCenterScale'),
    dict(type='RandomFlip', direction='horizontal'),
    dict(type='RandomHalfBody'),
    dict(
        type='RandomBBoxTransform',
        shift_factor=0.,
        scale_factor=[0.75, 1.25],
        rotate_factor=60),
    dict(type='TopdownAffine', input_size=codec['input_size']),
    dict(type='mmdet.YOLOXHSVRandomAug'),
    dict(
        type='Albumentation',
        transforms=[
            dict(type='Blur', p=0.1),
            dict(type='MedianBlur', p=0.1),
            dict(
                type='CoarseDropout',
                max_holes=1,
                max_height=0.4,
                max_width=0.4,
                min_holes=1,
                min_height=0.2,
                min_width=0.2,
                p=0.5),
        ]),
    dict(type='GenerateTarget', encoder=codec),
    dict(type='PackPoseInputs')
]

# data loaders
train_dataloader = dict(
    batch_size=256,
    num_workers=10,
    persistent_workers=True,
    sampler=dict(type='DefaultSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        data_mode=data_mode,
        ann_file='annotations/person_keypoints_train2017.json',
        data_prefix=dict(img='train2017/'),
        pipeline=train_pipeline,
    ))
val_dataloader = dict(
    batch_size=64,
    num_workers=10,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False, round_up=False),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        data_mode=data_mode,
        ann_file='annotations/person_keypoints_val2017.json',
        # bbox_file=f'{data_root}person_detection_results/'
        # 'COCO_val2017_detections_AP_H_56_person.json',
        data_prefix=dict(img='val2017/'),
        test_mode=True,
        pipeline=val_pipeline,
    ))
test_dataloader = val_dataloader

# hooks
default_hooks = dict(
    checkpoint=dict(save_best='coco/AP', rule='greater', max_keep_ckpts=1))

custom_hooks = [
    dict(
        type='EMAHook',
        ema_type='ExpMomentumEMA',
        momentum=0.0002,
        update_buffers=True,
        priority=49),
    dict(
        type='mmdet.PipelineSwitchHook',
        switch_epoch=max_epochs - stage2_num_epochs,
        switch_pipeline=train_pipeline_stage2)
]

# evaluators
val_evaluator = dict(
    type='CocoMetric',
    ann_file=data_root + 'annotations/person_keypoints_val2017.json')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.coders.tblr_bbox_coder.tblr2bboxes",
    backend="default",
)
def tblr2bboxes(
    priors, tblr, normalizer=4.0, normalize_by_wh=True, max_shape=None, clip_border=True
):
    """Rewrite `tblr2bboxes` for default backend.

    Since the need of clip op with dynamic min and max, this function uses
    clip_bboxes function to support dynamic shape.

    Args:
        ctx (ContextCaller): The context with additional information.
        priors (Tensor): Prior boxes in point form (x0, y0, x1, y1)
          Shape: (N,4) or (B, N, 4).
        tblr (Tensor): Coords of network output in tblr form
          Shape: (N, 4) or (B, N, 4).
        normalizer (Sequence[float] | float): Normalization parameter of
          encoded boxes. By list, it represents the normalization factors at
          tblr dims. By float, it is the unified normalization factor at all
          dims. Default: 4.0
        normalize_by_wh (bool): Whether the tblr coordinates have been
          normalized by the side length (wh) of prior bboxes.
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If priors shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]]
            and the length of max_shape should also be B.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.

    Return:
        bboxes (Tensor): Boxes with shape (N, 4) or (B, N, 4)
    """
    if not isinstance(normalizer, float):
        normalizer = torch.tensor(normalizer, device=priors.device)
        assert len(normalizer) == 4, "Normalizer must have length = 4"
    assert priors.size(0) == tblr.size(0)
    if priors.ndim == 3:
        assert priors.size(1) == tblr.size(1)

    loc_decode = tblr * normalizer
    prior_centers = (priors[..., 0:2] + priors[..., 2:4]) / 2
    if normalize_by_wh:
        wh = priors[..., 2:4] - priors[..., 0:2]

        w, h = torch.split(wh, 1, dim=-1)
        # Inplace operation with slice would fail for exporting to ONNX
        th = h * loc_decode[..., :2]  # tb
        tw = w * loc_decode[..., 2:]  # lr
        loc_decode = torch.cat([th, tw], dim=-1)
    top, bottom, left, right = loc_decode.split((1, 1, 1, 1), dim=-1)
    xmin = prior_centers[..., 0].unsqueeze(-1) - left
    xmax = prior_centers[..., 0].unsqueeze(-1) + right
    ymin = prior_centers[..., 1].unsqueeze(-1) - top
    ymax = prior_centers[..., 1].unsqueeze(-1) + bottom

    if clip_border and max_shape is not None:
        from juxtapose.mmdeploy.codebase.mmdet.deploy import clip_bboxes

        xmin, ymin, xmax, ymax = clip_bboxes(xmin, ymin, xmax, ymax, max_shape)
    bboxes = torch.cat([xmin, ymin, xmax, ymax], dim=-1).view(priors.size())

    return bboxes



# Copyright (c) OpenMMLab. All rights reserved.
from itertools import zip_longest
from typing import List, Optional, Sequence, Union

import mmengine
import torch
import torch.nn as nn
from mmengine import Config
from mmengine.model import BaseDataPreprocessor
from mmengine.registry import Registry
from mmengine.structures import BaseDataElement, InstanceData

from juxtapose.mmdeploy.codebase.base import BaseBackendModel
from juxtapose.mmdeploy.utils import (
    Backend,
    get_backend,
    get_codebase_config,
    load_config,
)

__BACKEND_MODEL = Registry("backend_segmentors")


@__BACKEND_MODEL.register_module("end2end")
class End2EndModel(BaseBackendModel):
    """End to end model for inference of pose detection.

    Args:
        backend (Backend): The backend enum, specifying backend type.
        backend_files (Sequence[str]): Paths to all required backend files(e.g.
            '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
        device (str): A string represents device type.
        deploy_cfg (str | mmengine.Config): Deployment config file or loaded
            Config object.
        model_cfg (str | mmengine.Config): Model config file or loaded Config
            object.
        data_preprocessor (dict | nn.Module | None): Input data pre-
                processor. Default is ``None``.
    """

    def __init__(
        self,
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        deploy_cfg: Union[str, mmengine.Config] = None,
        model_cfg: Union[str, mmengine.Config] = None,
        data_preprocessor: Optional[Union[dict, nn.Module]] = None,
        **kwargs,
    ):
        super(End2EndModel, self).__init__(
            deploy_cfg=deploy_cfg, data_preprocessor=data_preprocessor
        )
        from mmpose.models import builder

        self.deploy_cfg = deploy_cfg
        self.model_cfg = model_cfg
        self.device = device
        self._init_wrapper(
            backend=backend, backend_files=backend_files, device=device, **kwargs
        )
        # create head for decoding heatmap
        self.head = (
            builder.build_head(model_cfg.model.head)
            if hasattr(model_cfg.model, "head")
            else None
        )

    def _init_wrapper(
        self, backend: Backend, backend_files: Sequence[str], device: str, **kwargs
    ):
        """Initialize backend wrapper.

        Args:
            backend (Backend): The backend enum, specifying backend type.
            backend_files (Sequence[str]): Paths to all required backend files
                (e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
            device (str): A string specifying device type.
        """
        output_names = self.output_names
        self.wrapper = BaseBackendModel._build_wrapper(
            backend=backend,
            backend_files=backend_files,
            device=device,
            input_names=[self.input_name],
            output_names=output_names,
            deploy_cfg=self.deploy_cfg,
            **kwargs,
        )

    def forward(
        self,
        inputs: torch.Tensor,
        data_samples: List[BaseDataElement],
        mode: str = "predict",
        **kwargs,
    ):
        """Run forward inference.

        Args:
            inputs (torch.Tensor): Input image(s) in [N x C x H x W]
                format.
            data_samples (List[BaseDataElement]): A list of meta info for
                image(s).

        Returns:
            list: A list contains predictions.
        """
        assert mode == "predict", (
            "Backend model only support mode==predict," f" but get {mode}"
        )
        inputs = inputs.contiguous().to(self.device)
        batch_outputs = self.wrapper({self.input_name: inputs})
        batch_outputs = self.wrapper.output_to_list(batch_outputs)

        codebase_cfg = get_codebase_config(self.deploy_cfg)
        codec = self.model_cfg.codec
        if isinstance(codec, (list, tuple)):
            codec = codec[-1]

        if codec.type == "YOLOXPoseAnnotationProcessor":
            return self.pack_yolox_pose_result(batch_outputs, data_samples)
        elif codec.type == "SimCCLabel":
            export_postprocess = codebase_cfg.get("export_postprocess", False)
            if export_postprocess:
                keypoints, scores = [_.cpu().numpy() for _ in batch_outputs]
                preds = [InstanceData(keypoints=keypoints, keypoint_scores=scores)]
            else:
                batch_pred_x, batch_pred_y = batch_outputs
                preds = self.head.decode((batch_pred_x, batch_pred_y))
        elif codec.type in ["RegressionLabel", "IntegralRegressionLabel"]:
            preds = self.head.decode(batch_outputs)
        else:
            preds = self.head.decode(batch_outputs[0])
        results = self.pack_result(preds, data_samples)
        return results

    def pack_result(
        self,
        preds: Sequence[InstanceData],
        data_samples: List[BaseDataElement],
        convert_coordinate: bool = True,
    ):
        """Pack pred results to mmpose format
        Args:
            preds (Sequence[InstanceData]): Prediction of keypoints.
            data_samples (List[BaseDataElement]): A list of meta info for
                image(s).
            convert_coordinate (bool): Whether to convert keypoints
                coordinates to original image space. Default is True.
        Returns:
            data_samples (List[BaseDataElement]):
                updated data_samples with predictions.
        """
        if isinstance(preds, tuple):
            batch_pred_instances, batch_pred_fields = preds
        else:
            batch_pred_instances = preds
            batch_pred_fields = None
        assert len(batch_pred_instances) == len(data_samples)
        if batch_pred_fields is None:
            batch_pred_fields = []

        for pred_instances, pred_fields, data_sample in zip_longest(
            batch_pred_instances, batch_pred_fields, data_samples
        ):
            gt_instances = data_sample.gt_instances
            # convert keypoint coordinates from input space to image space
            if convert_coordinate:
                input_size = data_sample.metainfo["input_size"]
                input_center = data_sample.metainfo["input_center"]
                input_scale = data_sample.metainfo["input_scale"]
                keypoints = pred_instances.keypoints
                keypoints = keypoints / input_size * input_scale
                keypoints += input_center - 0.5 * input_scale
                pred_instances.keypoints = keypoints

            pred_instances.bboxes = gt_instances.bboxes
            pred_instances.bbox_scores = gt_instances.bbox_scores

            data_sample.pred_instances = pred_instances

            if pred_fields is not None:
                data_sample.pred_fields = pred_fields

        return data_samples

    def pack_yolox_pose_result(
        self, preds: List[torch.Tensor], data_samples: List[BaseDataElement]
    ):
        """Pack yolox-pose prediction results to mmpose format
        Args:
            preds (List[Tensor]): Prediction of bboxes and key-points.
            data_samples (List[BaseDataElement]): A list of meta info for
                image(s).
        Returns:
            data_samples (List[BaseDataElement]):
                updated data_samples with predictions.
        """
        assert preds[0].shape[0] == len(data_samples)
        batched_dets, batched_kpts = preds
        for data_sample_idx, data_sample in enumerate(data_samples):
            bboxes = batched_dets[data_sample_idx, :, :4]
            bbox_scores = batched_dets[data_sample_idx, :, 4]
            keypoints = batched_kpts[data_sample_idx, :, :, :2]
            keypoint_scores = batched_kpts[data_sample_idx, :, :, 2]

            # filter zero or negative scores
            inds = bbox_scores > 0.0
            bboxes = bboxes[inds, :]
            bbox_scores = bbox_scores[inds]
            keypoints = keypoints[inds, :]
            keypoint_scores = keypoint_scores[inds]

            pred_instances = InstanceData()

            # rescale
            input_size = data_sample.metainfo["input_size"]
            input_center = data_sample.metainfo["input_center"]
            input_scale = data_sample.metainfo["input_scale"]

            rescale = keypoints.new_tensor(input_scale) / keypoints.new_tensor(
                input_size
            )
            translation = keypoints.new_tensor(
                input_center
            ) - 0.5 * keypoints.new_tensor(input_scale)

            keypoints = keypoints * rescale.reshape(1, 1, 2) + translation.reshape(
                1, 1, 2
            )
            bboxes = bboxes * rescale.repeat(1, 2) + translation.repeat(1, 2)
            pred_instances.bboxes = bboxes.cpu().numpy()
            pred_instances.bbox_scores = bbox_scores
            # the precision test requires keypoints to be np.ndarray
            pred_instances.keypoints = keypoints.cpu().numpy()
            pred_instances.keypoint_scores = keypoint_scores
            pred_instances.lebels = torch.zeros(bboxes.shape[0])

            data_sample.pred_instances = pred_instances
        return data_samples


@__BACKEND_MODEL.register_module("sdk")
class SDKEnd2EndModel(End2EndModel):
    """SDK inference class, converts SDK output to mmpose format."""

    def __init__(self, *args, **kwargs):
        kwargs["data_preprocessor"] = None
        super(SDKEnd2EndModel, self).__init__(*args, **kwargs)
        self.ext_info = self.deploy_cfg.ext_info

    def forward(
        self,
        inputs: List[torch.Tensor],
        data_samples: List[BaseDataElement],
        mode: str = "predict",
        **kwargs,
    ) -> list:
        """Run forward inference.

        Args:
            inputs (List[torch.Tensor]): A list contains input image(s)
                in [H x W x C] format.
            data_samples (List[BaseDataElement]):
                Data samples of image metas.
            mode (str): test mode, only support 'predict'
            **kwargs: Other key-pair arguments.

        Returns:
            list: A list contains predictions.
        """
        pred_results = []
        for input_img, sample in zip(inputs, data_samples):
            bboxes = sample.gt_instances.bboxes

            # inputs are c,h,w, sdk requested h,w,c
            input_img = input_img.permute(1, 2, 0)
            input_img = input_img.contiguous().detach().cpu().numpy()
            keypoints = self.wrapper.handle(input_img, bboxes.tolist())
            pred = InstanceData(
                keypoints=keypoints[..., :2], keypoint_scores=keypoints[..., 2]
            )
            pred_results.append(pred)

        results = self.pack_result(pred_results, data_samples, convert_coordinate=False)
        return results


def build_pose_detection_model(
    model_files: Sequence[str],
    model_cfg: Union[str, mmengine.Config],
    deploy_cfg: Union[str, mmengine.Config],
    device: str,
    data_preprocessor: Optional[Union[Config, BaseDataPreprocessor]] = None,
    **kwargs,
):
    """Build object segmentation model for different backends.

    Args:
        model_files (Sequence[str]): Input model file(s).
        model_cfg (str | mmengine.Config): Input model config file or Config
            object.
        deploy_cfg (str | mmengine.Config): Input deployment config file or
            Config object.
        device (str):  Device to input model.
        data_preprocessor (Config | BaseDataPreprocessor | None): Input data
            pre-processor. Default is ``None``.
    Returns:
        BaseBackendModel: Pose model for a configured backend.
    """
    from mmpose.models.data_preprocessors import PoseDataPreprocessor

    # load cfg if necessary
    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

    backend = get_backend(deploy_cfg)
    model_type = get_codebase_config(deploy_cfg).get("model_type", "end2end")
    if isinstance(data_preprocessor, dict):
        dp = data_preprocessor.copy()
        dp_type = dp.pop("type")
        if dp_type == "mmdet.DetDataPreprocessor":
            from mmdet.models.data_preprocessors import DetDataPreprocessor

            data_preprocessor = DetDataPreprocessor(**dp)
        else:
            assert dp_type == "PoseDataPreprocessor"
            data_preprocessor = PoseDataPreprocessor(**dp)

    backend_pose_model = __BACKEND_MODEL.build(
        dict(
            type=model_type,
            backend=backend,
            backend_files=model_files,
            device=device,
            deploy_cfg=deploy_cfg,
            model_cfg=model_cfg,
            data_preprocessor=data_preprocessor,
            **kwargs,
        )
    )

    return backend_pose_model



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import get_common_config


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.csp_darknet.Focus.forward"
)
def focus__forward__default(self, x):
    """Rewrite forward function of Focus class.

    Replace slice with transpose.
    """
    # shape of x (b,c,w,h) -> y(b,4c,w/2,h/2)
    B, C, H, W = x.shape
    x = x.reshape(B, C, -1, 2, W)
    x = x.reshape(B, C, x.shape[2], 2, -1, 2)
    half_H = x.shape[2]
    half_W = x.shape[4]
    x = x.permute(0, 5, 3, 1, 2, 4)
    x = x.reshape(B, C * 4, half_H, half_W)

    return self.conv(x)


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.csp_darknet.Focus.forward", backend="ncnn"
)
def focus__forward__ncnn(self, x):
    """Rewrite forward function of Focus class for ncnn.

    Focus width and height information into channel space. ncnn does not
    support slice operator which step greater than 1, so we use another
    way to implement.

    Args:
        x (Tensor): The input tensor with shape (N, C, H, W).

    Returns:
        x (Tensor): The calculated tensor with shape (N, 4*C, H//2, W//2).
    """
    batch_size, c, h, w = x.shape
    assert (
        h % 2 == 0 and w % 2 == 0
    ), f"focus for yolox needs even feature\
        height and width, got {(h, w)}."

    x = x.reshape(batch_size, c * h, 1, w)
    _b, _c, _h, _w = x.shape
    g = torch.div(_c, 2, rounding_mode="floor")
    # fuse to ncnn's shufflechannel
    x = x.view(_b, g, 2, _h, _w)
    x = torch.transpose(x, 1, 2).contiguous()
    x = x.view(_b, -1, _h, _w)

    x = x.reshape(_b, c * h * w, 1, 1)

    _b, _c, _h, _w = x.shape
    g = torch.div(_c, 2, rounding_mode="floor")
    # fuse to ncnn's shufflechannel
    x = x.view(_b, g, 2, _h, _w)
    x = torch.transpose(x, 1, 2).contiguous()
    x = x.view(_b, -1, _h, _w)

    x = x.reshape(
        _b,
        c * 4,
        torch.div(h, 2, rounding_mode="floor"),
        torch.div(w, 2, rounding_mode="floor"),
    )

    return self.conv(x)


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.swin.WindowMSA.forward", backend="tensorrt"
)
def windowmsa__forward__tensorrt(self, x, mask=None):
    """Rewrite forward function of WindowMSA class for TensorRT.

    1. replace Gather operation of qkv with split.
    2. replace SoftMax operation with a workaround done by PyTorch.

    Args:
        x (tensor): input features with shape of (num_windows*B, N, C)
        mask (tensor | None, Optional): mask with shape of (num_windows,
            Wh*Ww, Wh*Ww), value should be between (-inf, 0].
    """
    ctx = FUNCTION_REWRITER.get_context()
    B, N, C = x.shape
    qkv = (
        self.qkv(x)
        .reshape(B, N, 3, self.num_heads, -1)
        .permute(2, 0, 3, 1, 4)
        .contiguous()
    )

    # replace the gather operation with the split
    q, k, v = [i.squeeze(0) for i in torch.split(qkv, 1, 0)]

    q = q * self.scale

    attn = q @ k.transpose(-2, -1)

    relative_position_bias = self.relative_position_bias_table[
        self.relative_position_index.view(-1)
    ].view(
        self.window_size[0] * self.window_size[1],
        self.window_size[0] * self.window_size[1],
        -1,
    )  # Wh*Ww,Wh*Ww,nH
    relative_position_bias = relative_position_bias.permute(
        2, 0, 1
    ).contiguous()  # nH, Wh*Ww, Wh*Ww
    attn = attn + relative_position_bias.unsqueeze(0)

    if mask is not None:
        nW = mask.shape[0]
        attn = attn.view(-1, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
        attn = attn.view(-1, self.num_heads, N, N)

    # replace softmax with a workaround
    # weird bug from TensorRT. softmax cannot be used here for fp32 and it
    # can be used in fp16, but softmax fp16 performance is not as good as
    # exp and log_softmax. Besides, only the UT of exp and log_softmax passed.
    fp16_mode = get_common_config(ctx.cfg).get("fp16_mode", False)
    if fp16_mode:
        attn = torch.exp(torch.log_softmax(attn, dim=self.softmax.dim))
    else:
        means = torch.mean(attn, self.softmax.dim, keepdim=True)[0]
        attn_exp = torch.exp(attn - means)
        attn_exp_sum = torch.sum(attn_exp, self.softmax.dim, keepdim=True)
        attn = attn_exp / attn_exp_sum

    attn = self.attn_drop(attn)

    x = (attn @ v).transpose(1, 2).contiguous().reshape(B, N, C)
    x = self.proj(x)
    x = self.proj_drop(x)
    return x


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.swin.ShiftWindowMSA.window_reverse",
    backend="tensorrt",
)
def shift_window_msa__window_reverse__tensorrt(self, windows, H, W):
    """Rewrite window_reverse function of ShiftWindowMSA class for TensorRT.
    For TensorRT, seems radical shape transformations are not allowed. Replace
    them with soft ones.

    Args:
        windows: (num_windows*B, window_size, window_size, C)
        H (int): Height of image
        W (int): Width of image
    Returns:
        x: (B, H, W, C)
    """
    window_size = self.window_size
    B = int(windows.shape[0] / (H * W / window_size / window_size))

    # x = windows.view(B, H // window_size, W // window_size, window_size,
    #     window_size, -1)
    x = windows.view(B, -1, W, window_size, windows.shape[-1])
    x = x.view(B, x.shape[1], -1, window_size, window_size, x.shape[-1])
    x = x.permute(0, 1, 3, 2, 4, 5).reshape(B, H, W, x.shape[-1])
    return x


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.swin.ShiftWindowMSA.window_partition",
    backend="tensorrt",
)
def shift_window_msa__window_partition__tensorrt(self, x):
    """Rewrite window_partition function of ShiftWindowMSA class for TensorRT.
    For TensorRT, seems radical shape transformations are not allowed. Replace
    them with soft ones.

    Args:
        x: (B, H, W, C)
    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    window_size = self.window_size
    x = x.view(B, H, -1, window_size, C)
    x = x.view(B, -1, window_size, x.shape[-3], window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous()
    windows = windows.view(-1, window_size, window_size, C)
    return windows


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.backbones.swin.ShiftWindowMSA.forward"
)
def shift_window_msa__forward__default(self, query, hw_shape):
    """Rewrite forward function of ShiftWindowMSA class.

    1. replace dynamic padding with static padding and dynamic slice.
    2. always do slice `x = x[:, :H, :W, :].contiguous()` for stability.
    """
    B, L, C = query.shape
    H, W = hw_shape
    assert L == H * W, "input feature has wrong size"
    query = query.view(B, H, W, C)

    # pad feature maps to multiples of window size
    query = query.permute(0, 3, 1, 2).contiguous()
    # query = torch.nn.ZeroPad2d([0, self.window_size, 0, self.window_size])(
    #     query)
    query = torch.cat([query, query.new_zeros(B, C, H, self.window_size)], dim=-1)
    query = torch.cat(
        [query, query.new_zeros(B, C, self.window_size, query.shape[-1])], dim=-2
    )
    slice_h = (
        torch.div((H + self.window_size - 1), self.window_size, rounding_mode="floor")
        * self.window_size
    )
    slice_w = (
        torch.div((W + self.window_size - 1), self.window_size, rounding_mode="floor")
        * self.window_size
    )
    query = query[:, :, :slice_h, :slice_w]
    query = query.permute(0, 2, 3, 1).contiguous()
    H_pad, W_pad = query.shape[1], query.shape[2]

    # cyclic shift
    if self.shift_size > 0:
        shifted_query = torch.roll(
            query, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2)
        )

        # calculate attention mask for SW-MSA
        w_mask = torch.cat(
            [
                shifted_query.new_zeros(W_pad - self.window_size),
                shifted_query.new_full((self.window_size - self.shift_size,), 1),
                shifted_query.new_full((self.shift_size,), 2),
            ]
        )
        h_mask = torch.cat(
            [
                shifted_query.new_zeros(H_pad - self.window_size),
                shifted_query.new_full((self.window_size - self.shift_size,), 3),
                shifted_query.new_full((self.shift_size,), 6),
            ]
        )

        img_mask = w_mask.unsqueeze(0) + h_mask.unsqueeze(1)
        img_mask = img_mask.unsqueeze(0)
        img_mask = img_mask.unsqueeze(-1)

        # nW, window_size, window_size, 1
        mask_windows = self.window_partition(img_mask)
        mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(
            attn_mask == 0, float(0.0)
        )
    else:
        shifted_query = query
        attn_mask = None

    # nW*B, window_size, window_size, C
    query_windows = self.window_partition(shifted_query)
    # nW*B, window_size*window_size, C
    query_windows = query_windows.view(-1, self.window_size**2, C)

    # W-MSA/SW-MSA (nW*B, window_size*window_size, C)
    attn_windows = self.w_msa(query_windows, mask=attn_mask)

    # merge windows
    attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)

    # B H' W' C
    shifted_x = self.window_reverse(attn_windows, H_pad, W_pad)
    # reverse cyclic shift
    if self.shift_size > 0:
        x = torch.roll(
            shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2)
        )
    else:
        x = shifted_x

    x = x[:, :H, :W, :].contiguous()

    x = x.view(B, H * W, C)

    x = self.drop(x)
    return x



# Copyright (c) OpenMMLab. All rights reserved.
from torch.onnx.symbolic_helper import parse_args

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_backend


@parse_args("v", "v", "i", "i", "i")
def grid_sampler(g, input, grid, interpolation_mode, padding_mode, align_corners=False):
    """Symbolic function for `grid_sampler`.

    PyTorch does not support export grid_sampler to ONNX by default. We add the
    support here. `grid_sampler` will be exported as ONNX node
    'mmdeploy::grid_sampler'
    """
    return g.op(
        "mmdeploy::grid_sampler",
        input,
        grid,
        interpolation_mode_i=interpolation_mode,
        padding_mode_i=padding_mode,
        align_corners_i=align_corners,
    )


@parse_args("v", "v", "i", "i", "i")
def grid_sampler_ppl(
    g, input, grid, interpolation_mode, padding_mode, align_corners=False
):
    """Symbolic function for `grid_sampler`.

    PyTorch does not support export grid_sampler to ONNX by default. We add the
    support here. `grid_sampler` will be exported as ONNX node
    'mmdeploy::grid_sampler'
    """
    return g.op(
        "mmcv::grid_sampler",
        input,
        grid,
        interpolation_mode_i=interpolation_mode,
        padding_mode_i=padding_mode,
        align_corners_i=align_corners,
    )


@SYMBOLIC_REWRITER.register_symbolic("grid_sampler", is_pytorch=True)
def grid_sampler__default(*args):
    """Register default symbolic function for `grid_sampler`.

    Add support to grid_sample to ONNX.
    """
    ctx = SYMBOLIC_REWRITER.get_context()
    backend = get_backend(ctx.cfg)
    if backend == Backend.PPLNN:
        return grid_sampler_ppl(*args)
    else:
        return grid_sampler(*args)



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import torch
from mmdet.models.detectors.base import ForwardResults
from mmdet.structures import DetDataSample
from mmdet.structures.det_data_sample import OptSampleList

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.utils import is_dynamic_shape


@mark("detector_forward", inputs=["input"], outputs=["dets", "labels", "masks"])
def __forward_impl(self, batch_inputs, data_samples):
    """Rewrite and adding mark for `forward`.

    Encapsulate this function for rewriting `forward` of BaseDetector.
    1. Add mark for BaseDetector.
    2. Support both dynamic and static export to onnx.
    """
    x = self.extract_feat(batch_inputs)

    output = self.bbox_head.predict(x, data_samples, rescale=False)
    return output


@torch.fx.wrap
def _set_metainfo(data_samples, img_shape):
    """Set the metainfo.

    Code in this function cannot be traced by fx.
    """

    # fx can not trace deepcopy correctly
    data_samples = copy.deepcopy(data_samples)
    if data_samples is None:
        data_samples = [DetDataSample()]

    # note that we can not use `set_metainfo`, deepcopy would crash the
    # onnx trace.
    for data_sample in data_samples:
        data_sample.set_field(name="img_shape", value=img_shape, field_type="metainfo")

    return data_samples


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.detectors.single_stage.SingleStageDetector.forward"
)
def single_stage_detector__forward(
    self,
    batch_inputs: torch.Tensor,
    data_samples: OptSampleList = None,
    mode: str = "tensor",
    **kwargs,
) -> ForwardResults:
    """Rewrite `forward` for default backend.

    Support configured dynamic/static shape for model input and return
    detection result as Tensor instead of numpy array.

    Args:
        batch_inputs (Tensor): Inputs with shape (N, C, H, W).
        data_samples (List[:obj:`DetDataSample`]): The Data
            Samples. It usually includes information such as
            `gt_instance`, `gt_panoptic_seg` and `gt_sem_seg`.
        mode (str): export mode, not used.

    Returns:
        tuple[Tensor]: Detection results of the
        input images.
            - dets (Tensor): Classification bboxes and scores.
                Has a shape (num_instances, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
    """
    ctx = FUNCTION_REWRITER.get_context()

    deploy_cfg = ctx.cfg

    # get origin input shape as tensor to support onnx dynamic shape
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    img_shape = torch._shape_as_tensor(batch_inputs)[2:]
    if not is_dynamic_flag:
        img_shape = [int(val) for val in img_shape]

    # set the metainfo
    data_samples = _set_metainfo(data_samples, img_shape)

    return __forward_impl(self, batch_inputs, data_samples=data_samples)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Tuple

import torch
import torch.nn.functional as F
from mmdet.structures.bbox import get_box_tensor
from mmengine import ConfigDict
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.bbox_heads.bbox_head.BBoxHead.forward"
)
@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.bbox_heads.convfc_bbox_head.ConvFCBBoxHead.forward"
)
def bbox_head__forward(self, x):
    """Rewrite `forward` for default backend.

    This function uses the specific `forward` function for the BBoxHead
    or ConvFCBBoxHead after adding marks.

    Args:
        ctx (ContextCaller): The context with additional information.
        self: The instance of the original class.
        x (Tensor): Input image tensor.

    Returns:
        tuple(Tensor, Tensor): The (cls_score, bbox_pred). The cls_score
        has shape (N, num_det, num_cls) and the bbox_pred has shape
        (N, num_det, 4).
    """
    ctx = FUNCTION_REWRITER.get_context()

    @mark(
        "bbox_head_forward", inputs=["bbox_feats"], outputs=["cls_score", "bbox_pred"]
    )
    def __forward(self, x):
        return ctx.origin_func(self, x)

    return __forward(self, x)


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.bbox_heads.bbox_head.BBoxHead.predict_by_feat"
)
def bbox_head__predict_by_feat(
    self,
    rois: Tuple[Tensor],
    cls_scores: Tuple[Tensor],
    bbox_preds: Tuple[Tensor],
    batch_img_metas: List[dict],
    rcnn_test_cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
) -> Tuple[Tensor]:
    """Rewrite `predict_by_feat` of `BBoxHead` for default backend.

    Transform network output for a batch into bbox predictions. Support
    `reg_class_agnostic == False` case.

    Args:
        rois (tuple[Tensor]): Tuple of boxes to be transformed.
            Each has shape  (num_boxes, 5). last dimension 5 arrange as
            (batch_index, x1, y1, x2, y2).
        cls_scores (tuple[Tensor]): Tuple of box scores, each has shape
            (num_boxes, num_classes + 1).
        bbox_preds (tuple[Tensor]): Tuple of box energies / deltas, each
            has shape (num_boxes, num_classes * 4).
        batch_img_metas (list[dict]): List of image information.
        rcnn_test_cfg (obj:`ConfigDict`, optional): `test_cfg` of R-CNN.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.

    Returns:
            - dets (Tensor): Classification bboxes and scores, has a shape
                (num_instance, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
    """
    ctx = FUNCTION_REWRITER.get_context()
    assert rois.ndim == 3, (
        "Only support export two stage " "model to ONNX " "with batch dimension. "
    )

    img_shape = batch_img_metas[0]["img_shape"]
    if self.custom_cls_channels:
        scores = self.loss_cls.get_activation(cls_scores)
    else:
        scores = F.softmax(cls_scores, dim=-1) if cls_scores is not None else None

    if bbox_preds is not None:
        # num_classes = 1 if self.reg_class_agnostic else self.num_classes
        # if num_classes > 1:
        #     rois = rois.repeat_interleave(num_classes, dim=1)
        bboxes = self.bbox_coder.decode(rois[..., 1:], bbox_preds, max_shape=img_shape)
        bboxes = get_box_tensor(bboxes)
    else:
        bboxes = rois[..., 1:].clone()
        if img_shape is not None:
            max_shape = bboxes.new_tensor(img_shape)[..., :2]
            min_xy = bboxes.new_tensor(0)
            max_xy = torch.cat([max_shape] * 2, dim=-1).flip(-1).unsqueeze(-2)
            bboxes = torch.where(bboxes < min_xy, min_xy, bboxes)
            bboxes = torch.where(bboxes > max_xy, max_xy, bboxes)

    batch_size = scores.shape[0]
    device = scores.device
    # ignore background class
    scores = scores[..., : self.num_classes]
    if not self.reg_class_agnostic:
        # only keep boxes with the max scores
        max_inds = scores.reshape(-1, self.num_classes).argmax(1, keepdim=True)
        encode_size = self.bbox_coder.encode_size
        bboxes = bboxes.reshape(-1, self.num_classes, encode_size)
        dim0_inds = torch.arange(bboxes.shape[0], device=device).unsqueeze(-1)
        bboxes = bboxes[dim0_inds, max_inds].reshape(batch_size, -1, encode_size)
    # get nms params
    post_params = get_post_processing_params(ctx.cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = rcnn_test_cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = rcnn_test_cfg.get("score_thr", post_params.score_threshold)
    if torch.onnx.is_in_onnx_export():
        pre_top_k = post_params.pre_top_k
    else:
        # For two stage partition post processing
        pre_top_k = (
            -1 if post_params.pre_top_k >= bboxes.shape[1] else post_params.pre_top_k
        )
    keep_top_k = rcnn_test_cfg.get("max_per_img", post_params.keep_top_k)
    nms_type = rcnn_test_cfg.nms.get("type")
    dets, labels = multiclass_nms(
        bboxes,
        scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )

    return dets, labels



# Copyright (c) OpenMMLab. All rights reserved.

import torch
from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend


class MultiHeadAttentionop(torch.autograd.Function):
    """Create onnx::MultiHeadAttention op."""

    @staticmethod
    def forward(
        ctx,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        q_weight: Tensor,
        q_bias: Tensor,
        k_weight: Tensor,
        k_bias: Tensor,
        v_weight: Tensor,
        v_bias: Tensor,
        o_weight: Tensor,
        o_bias: Tensor,
        embed_dims: int,
        num_heads: int,
    ) -> Tensor:
        return torch.rand_like(q)

    @staticmethod
    def symbolic(
        g,
        q: torch._C.Value,
        k: torch._C.Value,
        v: torch._C.Value,
        q_weight: torch._C.Value,
        q_bias: torch._C.Value,
        k_weight: torch._C.Value,
        k_bias: torch._C.Value,
        v_weight: torch._C.Value,
        v_bias: torch._C.Value,
        o_weight: torch._C.Value,
        o_bias: torch._C.Value,
        embed_dims: int,
        num_heads: int,
    ):
        q_weight.setDebugName("q_weight")
        q_bias.setDebugName("q_bias")

        k_weight.setDebugName("k_weight")
        k_bias.setDebugName("k_bias")

        v_weight.setDebugName("v_weight")
        v_bias.setDebugName("v_bias")

        o_weight.setDebugName("o_weight")
        o_bias.setDebugName("o_bias")

        return g.op(
            "mmdeploy::MultiHeadAttention",
            q,
            k,
            v,
            q_weight,
            q_bias,
            k_weight,
            k_bias,
            v_weight,
            v_bias,
            o_weight,
            o_bias,
            embed_dim_i=embed_dims,
            num_heads_i=num_heads,
        )


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmcv.cnn.bricks.transformer.MultiheadAttention.forward",
    backend=Backend.NCNN.value,
)
def multiheadattention__forward__ncnn(
    self,
    query,
    key=None,
    value=None,
    identity=None,
    query_pos=None,
    key_pos=None,
    attn_mask=None,
    key_padding_mask=None,
    **kwargs,
):
    """Rewrite `forward` of MultiheadAttention used in vision_transformer for
    ncnn backend.

    Args:
        query (Tensor): The input query with shape [num_queries, bs,
            embed_dims] if self.batch_first is False, else
            [bs, num_queries embed_dims].
        key (Tensor): The key tensor with shape [num_keys, bs,
            embed_dims] if self.batch_first is False, else
            [bs, num_keys, embed_dims] .
            If None, the ``query`` will be used. Defaults to None.
        value (Tensor): The value tensor with same shape as `key`.
            Same in `nn.MultiheadAttention.forward`. Defaults to None.
            If None, the `key` will be used.
        identity (Tensor): This tensor, with the same shape as x,
            will be used for the identity link.
            If None, `x` will be used. Defaults to None.
        query_pos (Tensor): The positional encoding for query, with
            the same shape as `x`. If not None, it will
            be added to `x` before forward function. Defaults to None.
        key_pos (Tensor): The positional encoding for `key`, with the
            same shape as `key`. Defaults to None. If not None, it will
            be added to `key` before forward function. If None, and
            `query_pos` has the same shape as `key`, then `query_pos`
            will be used for `key_pos`. Defaults to None.
        attn_mask (Tensor): ByteTensor mask with shape [num_queries,
            num_keys]. Same in `nn.MultiheadAttention.forward`.
            Defaults to None.
        key_padding_mask (Tensor): ByteTensor with shape [bs, num_keys].
            Defaults to None.
    Returns:
        Tensor: forwarded results with shape
        [bs, num_queries embed_dims].
    """
    if key is None:
        key = query
    if value is None:
        value = key
    if identity is None:
        identity = query
    if key_pos is None:
        if query_pos is not None:
            # use query_pos if key_pos is not available
            if query_pos.shape == key.shape:
                key_pos = query_pos
    if query_pos is not None:
        query = query + query_pos
    if key_pos is not None:
        key = key + key_pos

    assert query is key and key is value, "only support query==key==value"
    assert self.batch_first, "only support batch on first dim"
    assert attn_mask is None
    assert key_padding_mask is None

    # split qkv weight and bias
    qkv_weight = self.attn.in_proj_weight.data.reshape(3, -1, self.embed_dims)

    q_weight = qkv_weight[0]
    k_weight = qkv_weight[1]
    v_weight = qkv_weight[2]

    qkv_bias = self.attn.in_proj_bias.data.reshape(3, self.embed_dims)
    q_bias = qkv_bias[0]
    k_bias = qkv_bias[1]
    v_bias = qkv_bias[2]

    # out weight and bias
    o_weight = self.attn.out_proj.weight.data
    o_bias = self.attn.out_proj.bias.data
    # export to MultiHeadAttention in ncnn
    out = MultiHeadAttentionop.apply(
        query,
        key,
        value,
        q_weight,
        q_bias,
        k_weight,
        k_bias,
        v_weight,
        v_bias,
        o_weight,
        o_bias,
        self.embed_dims,
        self.num_heads,
    )
    return identity + self.dropout_layer(self.proj_drop(out))



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from mmengine import Config
from mmengine.dataset import pseudo_collate
from mmengine.model import BaseDataPreprocessor
from mmengine.registry import Registry

from juxtapose.mmdeploy.codebase.base import CODEBASE, BaseTask, MMCodebase
from juxtapose.mmdeploy.utils import Backend, Codebase, Task
from juxtapose.mmdeploy.utils.config_utils import (
    get_backend,
    get_input_shape,
    is_dynamic_shape,
)

MMDET_TASK = Registry("mmdet_tasks")


@CODEBASE.register_module(Codebase.MMDET.value)
class MMDetection(MMCodebase):
    """MMDetection codebase class."""

    task_registry = MMDET_TASK

    @classmethod
    def register_deploy_modules(cls):
        """register all rewriters for mmdet."""
        import juxtapose.mmdeploy.codebase.mmdet.models  # noqa: F401
        import juxtapose.mmdeploy.codebase.mmdet.ops
        import juxtapose.mmdeploy.codebase.mmdet.structures  # noqa: F401

    @classmethod
    def register_all_modules(cls):
        """register all related modules and rewriters for mmdet."""
        from mmdet.utils.setup_env import register_all_modules

        cls.register_deploy_modules()
        register_all_modules(True)


def process_model_config(
    model_cfg: Config,
    imgs: Union[Sequence[str], Sequence[np.ndarray]],
    input_shape: Optional[Sequence[int]] = None,
):
    """Process the model config.

    Args:
        model_cfg (Config): The model config.
        imgs (Sequence[str] | Sequence[np.ndarray]): Input image(s), accepted
            data type are List[str], List[np.ndarray].
        input_shape (list[int]): A list of two integer in (width, height)
            format specifying input shape. Default: None.

    Returns:
        Config: the model config after processing.
    """

    cfg = model_cfg.copy()

    if isinstance(imgs[0], np.ndarray):
        cfg = cfg.copy()
        # set loading pipeline type
        cfg.test_pipeline[0].type = "LoadImageFromNDArray"

    pipeline = cfg.test_pipeline

    for i, transform in enumerate(pipeline):
        # for static exporting
        if input_shape is not None:
            if transform.type == "Resize":
                pipeline[i].keep_ratio = False
                pipeline[i].scale = tuple(input_shape)
            if transform.type in ("YOLOv5KeepRatioResize", "LetterResize"):
                pipeline[i].scale = tuple(input_shape)

    pipeline = [
        transform for transform in pipeline if transform.type != "LoadAnnotations"
    ]
    cfg.test_pipeline = pipeline
    return cfg


def _get_dataset_metainfo(model_cfg: Config):
    """Get metainfo of dataset.

    Args:
        model_cfg Config: Input model Config object.

    Returns:
        list[str]: A list of string specifying names of different class.
    """
    from mmdet import datasets  # noqa
    from mmdet.registry import DATASETS

    module_dict = DATASETS.module_dict

    for dataloader_name in ["test_dataloader", "val_dataloader", "train_dataloader"]:
        if dataloader_name not in model_cfg:
            continue
        dataloader_cfg = model_cfg[dataloader_name]
        dataset_cfg = dataloader_cfg.dataset
        dataset_cls = module_dict.get(dataset_cfg.type, None)
        if dataset_cls is None:
            continue
        if hasattr(dataset_cls, "_load_metainfo") and isinstance(
            dataset_cls._load_metainfo, Callable
        ):
            meta = dataset_cls._load_metainfo(dataset_cfg.get("metainfo", None))
            if meta is not None:
                return meta
        if hasattr(dataset_cls, "METAINFO"):
            return dataset_cls.METAINFO

    return None


@MMDET_TASK.register_module(Task.OBJECT_DETECTION.value)
class ObjectDetection(BaseTask):
    """Object Detection task.

    Args:
        model_cfg (Config): The config of the model in mmdet.
        deploy_cfg (Config): The config of deployment.
        device (str): Device name.
    """

    def __init__(self, model_cfg: Config, deploy_cfg: Config, device: str) -> None:
        super().__init__(model_cfg, deploy_cfg, device)

        from mmcv.transforms import Compose

        cfg = model_cfg.copy()

        cfg.test_pipeline[0].type = "LoadImageFromNDArray"

        pipeline = [
            transform
            for transform in cfg.test_pipeline
            if transform.type != "LoadAnnotations"
        ]

        self.test_pipeline = Compose(pipeline)

    def build_backend_model(
        self,
        model_files: Optional[str] = None,
        data_preprocessor_updater: Optional[Callable] = None,
        **kwargs,
    ) -> torch.nn.Module:
        """Initialize backend model.

        Args:
            model_files (Sequence[str]): Input model files.
            data_preprocessor_updater (Callable | None): A function to update
                the data_preprocessor. Defaults to None.

        Returns:
            nn.Module: An initialized backend model.
        """
        from .object_detection_model import build_object_detection_model

        data_preprocessor = deepcopy(self.model_cfg.model.get("data_preprocessor", {}))
        if data_preprocessor_updater is not None:
            data_preprocessor = data_preprocessor_updater(data_preprocessor)
        data_preprocessor.setdefault("type", "mmdet.DetDataPreprocessor")

        model = build_object_detection_model(
            model_files,
            self.model_cfg,
            self.deploy_cfg,
            device=self.device,
            data_preprocessor=data_preprocessor,
        )
        model = model.to(self.device)
        return model.eval()

    def create_input(
        self,
        imgs: Union[str, np.ndarray],
        input_shape: Sequence[int] = None,
        data_preprocessor: Optional[BaseDataPreprocessor] = None,
    ) -> Tuple[Dict, torch.Tensor]:
        """Create input for detector.

        Args:
            imgs (str|np.ndarray): Input image(s), accpeted data type are
                `str`, `np.ndarray`.
            input_shape (list[int]): A list of two integer in (width, height)
                format specifying input shape. Defaults to `None`.
            data_preprocessor (BaseDataPreprocessor): The data preprocessor
                of the model. Default to `None`.

        Returns:
            tuple: (data, img), meta information for the input image and input.
        """

        if not isinstance(imgs, (list, tuple)):
            imgs = [imgs]

        data = []
        for img in imgs:
            # prepare data
            if isinstance(img, np.ndarray):
                # TODO: remove img_id.
                data_ = dict(img=img, img_id=0)
            else:
                # TODO: remove img_id.
                data_ = dict(img_path=img, img_id=0)
            # build the data pipeline
            data_ = self.test_pipeline(data_)
            data.append(data_)

        data = pseudo_collate(data)
        if data_preprocessor is not None:
            data = data_preprocessor(data, False)
            return data, data["inputs"]
        else:
            return data, BaseTask.get_tensor_from_input(data)

    @staticmethod
    def get_partition_cfg(partition_type: str) -> Dict:
        """Get a certain partition config for mmdet.

        Args:
            partition_type (str): A string specifying partition type.

        Returns:
            dict: A dictionary of partition config.
        """
        from .model_partition_cfg import MMDET_PARTITION_CFG

        assert (
            partition_type in MMDET_PARTITION_CFG
        ), f"Unknown partition_type {partition_type}"
        return MMDET_PARTITION_CFG[partition_type]

    def get_preprocess(self, *args, **kwargs) -> Dict:
        """Get the preprocess information for SDK.

        Return:
            dict: Composed of the preprocess information.
        """
        input_shape = get_input_shape(self.deploy_cfg)
        model_cfg = process_model_config(self.model_cfg, [""], input_shape)
        pipeline = model_cfg.test_pipeline
        meta_keys = [
            "filename",
            "ori_filename",
            "ori_shape",
            "img_shape",
            "pad_shape",
            "scale_factor",
            "flip",
            "flip_direction",
            "img_norm_cfg",
            "valid_ratio",
            "pad_param",
        ]
        # Extra pad outside datapreprocessor for CenterNet, CornerNet, etc.
        for i, transform in enumerate(pipeline):
            if transform["type"] == "RandomCenterCropPad":
                if transform["test_pad_mode"][0] == "logical_or":
                    extra_pad = dict(
                        type="Pad",
                        logical_or_val=transform["test_pad_mode"][1],
                        add_pix_val=transform["test_pad_add_pix"],
                    )
                    pipeline[i] = extra_pad
        transforms = [
            item
            for item in pipeline
            if "Random" not in item["type"] and "Annotation" not in item["type"]
        ]
        for i, transform in enumerate(transforms):
            # deal with mmyolo
            if transform["type"].startswith("mmdet."):
                transforms[i]["type"] = transform["type"][6:]
            if "PackDetInputs" in transform["type"]:
                meta_keys += transform["meta_keys"] if "meta_keys" in transform else []
                transform["meta_keys"] = list(set(meta_keys))
                transform["keys"] = ["img"]
                transforms[i]["type"] = "Collect"
            if transform["type"] == "Resize":
                transforms[i]["size"] = transforms[i].pop("scale")
        if self.codebase.value == "mmyolo":
            transforms = [
                item
                for item in pipeline
                if item["type"] not in ("ToGray", "YOLOv5KeepRatioResize")
            ]

        data_preprocessor = model_cfg.model.data_preprocessor

        transforms.insert(-1, dict(type="DefaultFormatBundle"))
        transforms.insert(
            -2,
            dict(type="Pad", size_divisor=data_preprocessor.get("pad_size_divisor", 1)),
        )
        transforms.insert(
            -3,
            dict(
                type="Normalize",
                to_rgb=data_preprocessor.get("bgr_to_rgb", False),
                mean=data_preprocessor.get("mean", [0, 0, 0]),
                std=data_preprocessor.get("std", [1, 1, 1]),
            ),
        )
        return transforms

    def get_postprocess(self, *args, **kwargs) -> Dict:
        """Get the postprocess information for SDK.

        Return:
            dict: Composed of the postprocess information.
        """
        params = self.model_cfg.model.test_cfg
        type = "ResizeBBox"  # default for object detection
        if "rpn" in params:
            params["min_bbox_size"] = params["rpn"]["min_bbox_size"]
        if "rcnn" in params:
            params["score_thr"] = params["rcnn"]["score_thr"]
            if "mask_thr_binary" in params["rcnn"]:
                params["mask_thr_binary"] = params["rcnn"]["mask_thr_binary"]
        if "mask_thr_binary" in params:
            type = "ResizeInstanceMask"  # for instance-seg
            params["is_resize_mask"] = False  # resize and crop mask default
        if get_backend(self.deploy_cfg) == Backend.RKNN:
            if (
                "YOLO" in self.model_cfg.model.type
                or "RTMDet" in self.model_cfg.model.type
            ):
                bbox_head = self.model_cfg.model.bbox_head
                type = bbox_head.type
                params["anchor_generator"] = bbox_head.get("anchor_generator", {})
                params["anchor_generator"].update(bbox_head.get("prior_generator", {}))
            else:  # default using base_dense_head
                type = "BaseDenseHead"
        return dict(type=type, params=params)

    def get_model_name(self, *args, **kwargs) -> str:
        """Get the model name.

        Return:
            str: the name of the model.
        """
        assert "type" in self.model_cfg.model, "model config contains no type"
        name = self.model_cfg.model.type.lower()
        return name

    def get_visualizer(self, name: str, save_dir: str):
        visualizer = super().get_visualizer(name, save_dir)
        metainfo = _get_dataset_metainfo(self.model_cfg)
        if metainfo is not None:
            visualizer.dataset_meta = metainfo
        return visualizer



# Copyright (c) OpenMMLab. All rights reserved.
import logging
import os.path as osp
from typing import Any, Callable, Optional, Sequence

from juxtapose.mmdeploy.utils import get_backend_config, get_common_config
from ..base import BACKEND_MANAGERS, BaseBackendManager


@BACKEND_MANAGERS.register("onnxruntime")
class ONNXRuntimeManager(BaseBackendManager):
    @classmethod
    def build_wrapper(
        cls,
        backend_files: Sequence[str],
        device: str = "cpu",
        input_names: Optional[Sequence[str]] = None,
        output_names: Optional[Sequence[str]] = None,
        deploy_cfg: Optional[Any] = None,
        **kwargs,
    ):
        """Build the wrapper for the backend model.

        Args:
            backend_files (Sequence[str]): Backend files.
            device (str, optional): The device info. Defaults to 'cpu'.
            input_names (Optional[Sequence[str]], optional): input names.
                Defaults to None.
            output_names (Optional[Sequence[str]], optional): output names.
                Defaults to None.
            deploy_cfg (Optional[Any], optional): The deploy config. Defaults
                to None.
        """

        from .wrapper import ORTWrapper

        return ORTWrapper(
            onnx_file=backend_files[0], device=device, output_names=output_names
        )

    @classmethod
    def is_available(cls, with_custom_ops: bool = False) -> bool:
        """Check whether backend is installed.

        Args:
            with_custom_ops (bool): check custom ops exists.
        Returns:
            bool: True if backend package is installed.
        """
        import importlib

        ret = importlib.util.find_spec("onnxruntime") is not None

        if ret and with_custom_ops:
            from .init_plugins import get_ops_path

            ops_path = get_ops_path()
            custom_ops_exist = osp.exists(ops_path)
            ret = ret and custom_ops_exist

        return ret

    @classmethod
    def get_version(cls) -> str:
        """Get the version of the backend."""
        if not cls.is_available():
            return "None"
        else:
            import pkg_resources

            try:
                ort_version = pkg_resources.get_distribution("onnxruntime").version
            except Exception:
                ort_version = "None"
            try:
                ort_gpu_version = pkg_resources.get_distribution(
                    "onnxruntime-gpu"
                ).version
            except Exception:
                ort_gpu_version = "None"

            if ort_gpu_version != "None":
                return ort_gpu_version
            else:
                return ort_version

    @classmethod
    def check_env(cls, log_callback: Callable = lambda _: _) -> str:
        """Check current environment.

        Returns:
            str: Info about the environment.
        """
        import pkg_resources

        try:
            if cls.is_available():
                ops_available = cls.is_available(with_custom_ops=True)
                ops_available = "Available" if ops_available else "NotAvailable"

                try:
                    ort_version = pkg_resources.get_distribution("onnxruntime").version
                except Exception:
                    ort_version = "None"
                try:
                    ort_gpu_version = pkg_resources.get_distribution(
                        "onnxruntime-gpu"
                    ).version
                except Exception:
                    ort_gpu_version = "None"

                ort_info = f"ONNXRuntime:\t{ort_version}"
                log_callback(ort_info)
                ort_gpu_info = f"ONNXRuntime-gpu:\t{ort_gpu_version}"
                log_callback(ort_gpu_info)
                ort_ops_info = f"ONNXRuntime custom ops:\t{ops_available}"
                log_callback(ort_ops_info)

                info = f"{ort_info}\n{ort_gpu_info}\n{ort_ops_info}"
            else:
                info = "ONNXRuntime:\tNone"
                log_callback(info)
        except Exception:
            info = f"{cls.backend_name}:\tCheckFailed"
            log_callback(info)
        return info

    @classmethod
    def to_backend(
        cls,
        ir_files: Sequence[str],
        work_dir: str,
        deploy_cfg: Any,
        log_level: int = logging.INFO,
        device: str = "cpu",
        **kwargs,
    ) -> Sequence[str]:
        """Convert intermediate representation to given backend.

        Args:
            ir_files (Sequence[str]): The intermediate representation files.
            work_dir (str): The work directory, backend files and logs should
                be saved in this directory.
            deploy_cfg (Any): The deploy config.
            log_level (int, optional): The log level. Defaults to logging.INFO.
            device (str, optional): The device type. Defaults to 'cpu'.
        Returns:
            Sequence[str]: Backend files.
        """
        backend_cfg = get_backend_config(deploy_cfg)

        precision = backend_cfg.get("precision", "fp32")
        if precision == "fp16":
            import onnx
            from onnxconverter_common import float16

            common_cfg = get_common_config(deploy_cfg)
            model = onnx.load(ir_files[0])
            model_fp16 = float16.convert_float_to_float16(model, **common_cfg)
            onnx.save(model_fp16, ir_files[0])
        return ir_files



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Union

import torch
from torch.types import Number

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils.constants import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.masked_fill", backend=Backend.ONNXRUNTIME.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.masked_fill", backend=Backend.ONNXRUNTIME.value
)
def masked_fill__onnxruntime(
    input, mask: torch.Tensor, value: Union[torch.Tensor, Number]
) -> torch.Tensor:
    """Rewrite `masked_fill` for onnxruntime backend.

    SATRN model as example, when value is set to `float('-inf')`, the results
    of ORT inferencing turns out to be NAN.
    """
    ctx = FUNCTION_REWRITER.get_context()
    if value == float("-inf"):
        value = -1e34  # hard coding number
    return ctx.origin_func(input, mask, value)



# Copyright (c) OpenMMLab. All rights reserved.
from .backend_manager import (BACKEND_MANAGERS, BaseBackendManager,
                              get_backend_manager)
from .backend_wrapper_registry import (BACKEND_WRAPPER, get_backend_file_count,
                                       get_backend_wrapper_class)
from .base_wrapper import BaseWrapper

__all__ = [
    'BACKEND_MANAGERS', 'BaseBackendManager', 'get_backend_manager',
    'BaseWrapper', 'BACKEND_WRAPPER', 'get_backend_wrapper_class',
    'get_backend_file_count'
]



# Copyright (c) OpenMMLab. All rights reserved.
import mmdet.structures.bbox.transforms

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.coders.distance_point_bbox_coder"
    ".DistancePointBBoxCoder.decode",
    backend="default",
)
def distancepointbboxcoder__decode(self, points, pred_bboxes, max_shape=None):
    """Rewrite `mmdet.models.task_modules.coders.distance_point_bbox_coder. \
    DistancePointBBoxCoder.decode`

    Decode distance prediction to bounding box.

    Args:
        ctx (ContextCaller): The context with additional information.
        self (DistancePointBBoxCoder): The instance of the class
            DistancePointBBoxCoder.
        points (Tensor): Shape (B, N, 2) or (N, 2).
        pred_bboxes (Tensor): Distance from the given point to 4
            boundaries (left, top, right, bottom). Shape (B, N, 4)
            or (N, 4)
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If priors shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]],
            and the length of max_shape should also be B.
            Default None.
    Returns:
        Tensor: Boxes with shape (N, 4) or (B, N, 4)
    """
    assert points.size(0) == pred_bboxes.size(0)
    assert points.size(-1) == 2
    assert pred_bboxes.size(-1) == 4
    if self.clip_border is False:
        max_shape = None
    # Rewrite add mmdet.core.bbox.transforms to find correct
    # rewriter, or you will not find correct rewriter.
    return mmdet.structures.bbox.transforms.distance2bbox(
        points, pred_bboxes, max_shape
    )



# Copyright (c) OpenMMLab. All rights reserved.
# modified from
# https://github.com/pytorch/pytorch/blob/master/torch/onnx/symbolic_opset9.py
import sys

from torch.onnx.symbolic_helper import _slice_helper, parse_args

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@parse_args("v", "is", "is")
def roll(g, self, shifts, dims):
    """Symbolic function for `roll`."""
    assert len(shifts) == len(dims)

    result = self
    for i in range(len(shifts)):
        shapes = []
        shape = _slice_helper(
            g, result, axes=[dims[i]], starts=[-shifts[i]], ends=[sys.maxsize]
        )
        shapes.append(shape)
        shape = _slice_helper(g, result, axes=[dims[i]], starts=[0], ends=[-shifts[i]])
        shapes.append(shape)
        result = g.op("Concat", *shapes, axis_i=dims[i])

    return result


@SYMBOLIC_REWRITER.register_symbolic("roll", is_pytorch=True)
def roll_default(g, self, shifts, dims):
    """Support export roll to ONNX with PyTorch version 1.10-."""
    return roll(g, self, shifts, dims)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Sequence, Union

import mmengine
import numpy as np


def inference_model(
    model_cfg: Union[str, mmengine.Config],
    deploy_cfg: Union[str, mmengine.Config],
    backend_files: Sequence[str],
    img: Union[str, np.ndarray],
    device: str,
) -> Any:
    """Run inference with PyTorch or backend model and show results.

    Examples:
        >>> from juxtapose.mmdeploy.apis import inference_model
        >>> model_cfg = ('mmdetection/configs/fcos/'
                         'fcos_r50_caffe_fpn_gn-head_1x_coco.py')
        >>> deploy_cfg = ('configs/mmdet/detection/'
                          'detection_onnxruntime_dynamic.py')
        >>> backend_files = ['work_dir/fcos.onnx']
        >>> img = 'demo.jpg'
        >>> device = 'cpu'
        >>> model_output = inference_model(model_cfg, deploy_cfg,
                            backend_files, img, device)

    Args:
        model_cfg (str | mmengine.Config): Model config file or Config object.
        deploy_cfg (str | mmengine.Config): Deployment config file or Config
            object.
        backend_files (Sequence[str]): Input backend model file(s).
        img (str | np.ndarray): Input image file or numpy array for inference.
        device (str): A string specifying device type.

    Returns:
        Any: The inference results
    """
    import torch

    from juxtapose.mmdeploy.utils import get_input_shape, load_config

    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

    from juxtapose.mmdeploy.apis.utils import build_task_processor

    task_processor = build_task_processor(model_cfg, deploy_cfg, device)

    model = task_processor.build_backend_model(
        backend_files, task_processor.update_data_preprocessor
    )

    input_shape = get_input_shape(deploy_cfg)
    model_inputs, _ = task_processor.create_input(img, input_shape)

    with torch.no_grad():
        result = model.test_step(model_inputs)

    return result



# Copyright (c) OpenMMLab. All rights reserved.
import logging
from typing import Optional

logger_initialized = {}


def get_logger(name: str,
               log_file: Optional[str] = None,
               log_level: int = logging.INFO,
               file_mode: str = 'w'):
    """Initialize and get a logger by name.

    If the logger has not been initialized, this method will initialize the
    logger by adding one or two handlers, otherwise the initialized logger will
    be directly returned. During initialization, a StreamHandler will always be
    added. If `log_file` is specified, a FileHandler will also be added.
    Args:
        name (str): Logger name.
        log_file (str | None): The log filename. If specified, a FileHandler
            will be added to the logger.
        log_level (int): The logger level.
        file_mode (str): The file mode used in opening log file.
            Defaults to 'w'.
    Returns:
        logging.Logger: The expected logger.
    """
    # use logger in mmengine if exist.
    try:
        from mmengine.logging import MMLogger
        if MMLogger.check_instance_created(name):
            logger = MMLogger.get_instance(name)
        else:
            logger = MMLogger.get_instance(
                name,
                log_file=log_file,
                log_level=log_level,
                file_mode=file_mode)
        return logger

    except Exception:
        pass

    logger = logging.getLogger(name)
    if name in logger_initialized:
        return logger
    # handle hierarchical names
    # e.g., logger "a" is initialized, then logger "a.b" will skip the
    # initialization since it is a child of "a".
    for logger_name in logger_initialized:
        if name.startswith(logger_name):
            return logger

    # handle duplicate logs to the console
    for handler in logger.root.handlers:
        if type(handler) is logging.StreamHandler:
            handler.setLevel(logging.ERROR)

    stream_handler = logging.StreamHandler()
    handlers = [stream_handler]

    if log_file is not None:
        # Here, the default behaviour of the official logger is 'a'. Thus, we
        # provide an interface to change the file mode to the default
        # behaviour.
        file_handler = logging.FileHandler(log_file, file_mode)
        handlers.append(file_handler)

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    for handler in handlers:
        handler.setFormatter(formatter)
        handler.setLevel(log_level)
        logger.addHandler(handler)

    logger.setLevel(log_level)
    logger_initialized[name] = True

    return logger



from .bot_sort import BOTSORT
from .byte_tracker import BYTETracker
from .track import Tracker, TRACKER_MAP


# from .track import register_tracker

__all__ = "Tracker", "TRACKER_MAP", "BOTSORT", "BYTETracker"  # allow simpler import
